/* DO NOT EDIT - This file generated automatically by api_vtxfmt_init_h.py script */

/*
 * Copyright (C) 2012 Intel Corporation
 * Copyright (C) 2021 Advanced Micro Devices, Inc.
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  IN NO EVENT SHALL
 * AUTHORS,
 * AND/OR THEIR SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
 * OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

if (_mesa_is_desktop_gl(ctx)) {
   SET_VertexAttrib1d(tab, NAME(VertexAttrib1d));
   SET_VertexAttrib1dv(tab, NAME(VertexAttrib1dv));
   SET_VertexAttrib1fARB(tab, NAME(VertexAttrib1fARB));
   SET_VertexAttrib1fvARB(tab, NAME(VertexAttrib1fvARB));
   SET_VertexAttrib1s(tab, NAME(VertexAttrib1s));
   SET_VertexAttrib1sv(tab, NAME(VertexAttrib1sv));
   SET_VertexAttrib2d(tab, NAME(VertexAttrib2d));
   SET_VertexAttrib2dv(tab, NAME(VertexAttrib2dv));
   SET_VertexAttrib2fARB(tab, NAME(VertexAttrib2fARB));
   SET_VertexAttrib2fvARB(tab, NAME(VertexAttrib2fvARB));
   SET_VertexAttrib2s(tab, NAME(VertexAttrib2s));
   SET_VertexAttrib2sv(tab, NAME(VertexAttrib2sv));
   SET_VertexAttrib3d(tab, NAME(VertexAttrib3d));
   SET_VertexAttrib3dv(tab, NAME(VertexAttrib3dv));
   SET_VertexAttrib3fARB(tab, NAME(VertexAttrib3fARB));
   SET_VertexAttrib3fvARB(tab, NAME(VertexAttrib3fvARB));
   SET_VertexAttrib3s(tab, NAME(VertexAttrib3s));
   SET_VertexAttrib3sv(tab, NAME(VertexAttrib3sv));
   SET_VertexAttrib4Nbv(tab, NAME(VertexAttrib4Nbv));
   SET_VertexAttrib4Niv(tab, NAME(VertexAttrib4Niv));
   SET_VertexAttrib4Nsv(tab, NAME(VertexAttrib4Nsv));
   SET_VertexAttrib4Nub(tab, NAME(VertexAttrib4Nub));
   SET_VertexAttrib4Nubv(tab, NAME(VertexAttrib4Nubv));
   SET_VertexAttrib4Nuiv(tab, NAME(VertexAttrib4Nuiv));
   SET_VertexAttrib4Nusv(tab, NAME(VertexAttrib4Nusv));
   SET_VertexAttrib4bv(tab, NAME(VertexAttrib4bv));
   SET_VertexAttrib4d(tab, NAME(VertexAttrib4d));
   SET_VertexAttrib4dv(tab, NAME(VertexAttrib4dv));
   SET_VertexAttrib4fARB(tab, NAME(VertexAttrib4fARB));
   SET_VertexAttrib4fvARB(tab, NAME(VertexAttrib4fvARB));
   SET_VertexAttrib4iv(tab, NAME(VertexAttrib4iv));
   SET_VertexAttrib4s(tab, NAME(VertexAttrib4s));
   SET_VertexAttrib4sv(tab, NAME(VertexAttrib4sv));
   SET_VertexAttrib4ubv(tab, NAME(VertexAttrib4ubv));
   SET_VertexAttrib4uiv(tab, NAME(VertexAttrib4uiv));
   SET_VertexAttrib4usv(tab, NAME(VertexAttrib4usv));
   SET_VertexAttribI1iEXT(tab, NAME(VertexAttribI1iEXT));
   SET_VertexAttribI1iv(tab, NAME(VertexAttribI1iv));
   SET_VertexAttribI1uiEXT(tab, NAME(VertexAttribI1uiEXT));
   SET_VertexAttribI1uiv(tab, NAME(VertexAttribI1uiv));
   SET_VertexAttribI2iEXT(tab, NAME(VertexAttribI2iEXT));
   SET_VertexAttribI2ivEXT(tab, NAME(VertexAttribI2ivEXT));
   SET_VertexAttribI2uiEXT(tab, NAME(VertexAttribI2uiEXT));
   SET_VertexAttribI2uivEXT(tab, NAME(VertexAttribI2uivEXT));
   SET_VertexAttribI3iEXT(tab, NAME(VertexAttribI3iEXT));
   SET_VertexAttribI3ivEXT(tab, NAME(VertexAttribI3ivEXT));
   SET_VertexAttribI3uiEXT(tab, NAME(VertexAttribI3uiEXT));
   SET_VertexAttribI3uivEXT(tab, NAME(VertexAttribI3uivEXT));
   SET_VertexAttribI4bv(tab, NAME(VertexAttribI4bv));
   SET_VertexAttribI4sv(tab, NAME(VertexAttribI4sv));
   SET_VertexAttribI4ubv(tab, NAME(VertexAttribI4ubv));
   SET_VertexAttribI4usv(tab, NAME(VertexAttribI4usv));
   SET_VertexAttribL1ui64ARB(tab, NAME(VertexAttribL1ui64ARB));
   SET_VertexAttribL1ui64vARB(tab, NAME(VertexAttribL1ui64vARB));
   SET_VertexAttribP1ui(tab, NAME(VertexAttribP1ui));
   SET_VertexAttribP1uiv(tab, NAME(VertexAttribP1uiv));
   SET_VertexAttribP2ui(tab, NAME(VertexAttribP2ui));
   SET_VertexAttribP2uiv(tab, NAME(VertexAttribP2uiv));
   SET_VertexAttribP3ui(tab, NAME(VertexAttribP3ui));
   SET_VertexAttribP3uiv(tab, NAME(VertexAttribP3uiv));
   SET_VertexAttribP4ui(tab, NAME(VertexAttribP4ui));
   SET_VertexAttribP4uiv(tab, NAME(VertexAttribP4uiv));
}
if (_mesa_is_desktop_gl(ctx) || (ctx->API == API_OPENGLES2 && ctx->Version >= 30)) {
   SET_VertexAttribI4iEXT(tab, NAME(VertexAttribI4iEXT));
   SET_VertexAttribI4ivEXT(tab, NAME(VertexAttribI4ivEXT));
   SET_VertexAttribI4uiEXT(tab, NAME(VertexAttribI4uiEXT));
   SET_VertexAttribI4uivEXT(tab, NAME(VertexAttribI4uivEXT));
}
if (ctx->API == API_OPENGLES2) {
   SET_VertexAttrib1fARB(tab, NAME_ES(VertexAttrib1fARB));
   SET_VertexAttrib1fvARB(tab, NAME_ES(VertexAttrib1fvARB));
   SET_VertexAttrib2fARB(tab, NAME_ES(VertexAttrib2fARB));
   SET_VertexAttrib2fvARB(tab, NAME_ES(VertexAttrib2fvARB));
   SET_VertexAttrib3fARB(tab, NAME_ES(VertexAttrib3fARB));
   SET_VertexAttrib3fvARB(tab, NAME_ES(VertexAttrib3fvARB));
   SET_VertexAttrib4fARB(tab, NAME_ES(VertexAttrib4fARB));
   SET_VertexAttrib4fvARB(tab, NAME_ES(VertexAttrib4fvARB));
}
if (ctx->API == API_OPENGL_COMPAT) {
   SET_ArrayElement(tab, NAME_AE(ArrayElement));
   SET_Begin(tab, NAME(Begin));
   SET_CallList(tab, NAME_CALLLIST(CallList));
   SET_CallLists(tab, NAME_CALLLIST(CallLists));
   SET_Color3b(tab, NAME(Color3b));
   SET_Color3bv(tab, NAME(Color3bv));
   SET_Color3d(tab, NAME(Color3d));
   SET_Color3dv(tab, NAME(Color3dv));
   SET_Color3f(tab, NAME(Color3f));
   SET_Color3fv(tab, NAME(Color3fv));
   SET_Color3hNV(tab, NAME(Color3hNV));
   SET_Color3hvNV(tab, NAME(Color3hvNV));
   SET_Color3i(tab, NAME(Color3i));
   SET_Color3iv(tab, NAME(Color3iv));
   SET_Color3s(tab, NAME(Color3s));
   SET_Color3sv(tab, NAME(Color3sv));
   SET_Color3ub(tab, NAME(Color3ub));
   SET_Color3ubv(tab, NAME(Color3ubv));
   SET_Color3ui(tab, NAME(Color3ui));
   SET_Color3uiv(tab, NAME(Color3uiv));
   SET_Color3us(tab, NAME(Color3us));
   SET_Color3usv(tab, NAME(Color3usv));
   SET_Color4b(tab, NAME(Color4b));
   SET_Color4bv(tab, NAME(Color4bv));
   SET_Color4d(tab, NAME(Color4d));
   SET_Color4dv(tab, NAME(Color4dv));
   SET_Color4fv(tab, NAME(Color4fv));
   SET_Color4hNV(tab, NAME(Color4hNV));
   SET_Color4hvNV(tab, NAME(Color4hvNV));
   SET_Color4i(tab, NAME(Color4i));
   SET_Color4iv(tab, NAME(Color4iv));
   SET_Color4s(tab, NAME(Color4s));
   SET_Color4sv(tab, NAME(Color4sv));
   SET_Color4ubv(tab, NAME(Color4ubv));
   SET_Color4ui(tab, NAME(Color4ui));
   SET_Color4uiv(tab, NAME(Color4uiv));
   SET_Color4us(tab, NAME(Color4us));
   SET_Color4usv(tab, NAME(Color4usv));
   SET_ColorP3ui(tab, NAME(ColorP3ui));
   SET_ColorP3uiv(tab, NAME(ColorP3uiv));
   SET_ColorP4ui(tab, NAME(ColorP4ui));
   SET_ColorP4uiv(tab, NAME(ColorP4uiv));
   SET_EdgeFlag(tab, NAME(EdgeFlag));
   SET_EdgeFlagv(tab, NAME(EdgeFlagv));
   SET_End(tab, NAME(End));
   SET_EvalCoord1d(tab, NAME(EvalCoord1d));
   SET_EvalCoord1dv(tab, NAME(EvalCoord1dv));
   SET_EvalCoord1f(tab, NAME(EvalCoord1f));
   SET_EvalCoord1fv(tab, NAME(EvalCoord1fv));
   SET_EvalCoord2d(tab, NAME(EvalCoord2d));
   SET_EvalCoord2dv(tab, NAME(EvalCoord2dv));
   SET_EvalCoord2f(tab, NAME(EvalCoord2f));
   SET_EvalCoord2fv(tab, NAME(EvalCoord2fv));
   SET_EvalPoint1(tab, NAME(EvalPoint1));
   SET_EvalPoint2(tab, NAME(EvalPoint2));
   SET_FogCoordd(tab, NAME(FogCoordd));
   SET_FogCoorddv(tab, NAME(FogCoorddv));
   SET_FogCoordfEXT(tab, NAME(FogCoordfEXT));
   SET_FogCoordfvEXT(tab, NAME(FogCoordfvEXT));
   SET_FogCoordhNV(tab, NAME(FogCoordhNV));
   SET_FogCoordhvNV(tab, NAME(FogCoordhvNV));
   SET_Indexd(tab, NAME(Indexd));
   SET_Indexdv(tab, NAME(Indexdv));
   SET_Indexf(tab, NAME(Indexf));
   SET_Indexfv(tab, NAME(Indexfv));
   SET_Indexi(tab, NAME(Indexi));
   SET_Indexiv(tab, NAME(Indexiv));
   SET_Indexs(tab, NAME(Indexs));
   SET_Indexsv(tab, NAME(Indexsv));
   SET_Indexub(tab, NAME(Indexub));
   SET_Indexubv(tab, NAME(Indexubv));
   SET_Materiali(tab, NAME(Materiali));
   SET_Materialiv(tab, NAME(Materialiv));
   SET_MultiTexCoord1d(tab, NAME(MultiTexCoord1d));
   SET_MultiTexCoord1dv(tab, NAME(MultiTexCoord1dv));
   SET_MultiTexCoord1fARB(tab, NAME(MultiTexCoord1fARB));
   SET_MultiTexCoord1fvARB(tab, NAME(MultiTexCoord1fvARB));
   SET_MultiTexCoord1hNV(tab, NAME(MultiTexCoord1hNV));
   SET_MultiTexCoord1hvNV(tab, NAME(MultiTexCoord1hvNV));
   SET_MultiTexCoord1i(tab, NAME(MultiTexCoord1i));
   SET_MultiTexCoord1iv(tab, NAME(MultiTexCoord1iv));
   SET_MultiTexCoord1s(tab, NAME(MultiTexCoord1s));
   SET_MultiTexCoord1sv(tab, NAME(MultiTexCoord1sv));
   SET_MultiTexCoord2d(tab, NAME(MultiTexCoord2d));
   SET_MultiTexCoord2dv(tab, NAME(MultiTexCoord2dv));
   SET_MultiTexCoord2fARB(tab, NAME(MultiTexCoord2fARB));
   SET_MultiTexCoord2fvARB(tab, NAME(MultiTexCoord2fvARB));
   SET_MultiTexCoord2hNV(tab, NAME(MultiTexCoord2hNV));
   SET_MultiTexCoord2hvNV(tab, NAME(MultiTexCoord2hvNV));
   SET_MultiTexCoord2i(tab, NAME(MultiTexCoord2i));
   SET_MultiTexCoord2iv(tab, NAME(MultiTexCoord2iv));
   SET_MultiTexCoord2s(tab, NAME(MultiTexCoord2s));
   SET_MultiTexCoord2sv(tab, NAME(MultiTexCoord2sv));
   SET_MultiTexCoord3d(tab, NAME(MultiTexCoord3d));
   SET_MultiTexCoord3dv(tab, NAME(MultiTexCoord3dv));
   SET_MultiTexCoord3fARB(tab, NAME(MultiTexCoord3fARB));
   SET_MultiTexCoord3fvARB(tab, NAME(MultiTexCoord3fvARB));
   SET_MultiTexCoord3hNV(tab, NAME(MultiTexCoord3hNV));
   SET_MultiTexCoord3hvNV(tab, NAME(MultiTexCoord3hvNV));
   SET_MultiTexCoord3i(tab, NAME(MultiTexCoord3i));
   SET_MultiTexCoord3iv(tab, NAME(MultiTexCoord3iv));
   SET_MultiTexCoord3s(tab, NAME(MultiTexCoord3s));
   SET_MultiTexCoord3sv(tab, NAME(MultiTexCoord3sv));
   SET_MultiTexCoord4d(tab, NAME(MultiTexCoord4d));
   SET_MultiTexCoord4dv(tab, NAME(MultiTexCoord4dv));
   SET_MultiTexCoord4fvARB(tab, NAME(MultiTexCoord4fvARB));
   SET_MultiTexCoord4hNV(tab, NAME(MultiTexCoord4hNV));
   SET_MultiTexCoord4hvNV(tab, NAME(MultiTexCoord4hvNV));
   SET_MultiTexCoord4i(tab, NAME(MultiTexCoord4i));
   SET_MultiTexCoord4iv(tab, NAME(MultiTexCoord4iv));
   SET_MultiTexCoord4s(tab, NAME(MultiTexCoord4s));
   SET_MultiTexCoord4sv(tab, NAME(MultiTexCoord4sv));
   SET_MultiTexCoordP1ui(tab, NAME(MultiTexCoordP1ui));
   SET_MultiTexCoordP1uiv(tab, NAME(MultiTexCoordP1uiv));
   SET_MultiTexCoordP2ui(tab, NAME(MultiTexCoordP2ui));
   SET_MultiTexCoordP2uiv(tab, NAME(MultiTexCoordP2uiv));
   SET_MultiTexCoordP3ui(tab, NAME(MultiTexCoordP3ui));
   SET_MultiTexCoordP3uiv(tab, NAME(MultiTexCoordP3uiv));
   SET_MultiTexCoordP4ui(tab, NAME(MultiTexCoordP4ui));
   SET_MultiTexCoordP4uiv(tab, NAME(MultiTexCoordP4uiv));
   SET_Normal3b(tab, NAME(Normal3b));
   SET_Normal3bv(tab, NAME(Normal3bv));
   SET_Normal3d(tab, NAME(Normal3d));
   SET_Normal3dv(tab, NAME(Normal3dv));
   SET_Normal3fv(tab, NAME(Normal3fv));
   SET_Normal3hNV(tab, NAME(Normal3hNV));
   SET_Normal3hvNV(tab, NAME(Normal3hvNV));
   SET_Normal3i(tab, NAME(Normal3i));
   SET_Normal3iv(tab, NAME(Normal3iv));
   SET_Normal3s(tab, NAME(Normal3s));
   SET_Normal3sv(tab, NAME(Normal3sv));
   SET_NormalP3ui(tab, NAME(NormalP3ui));
   SET_NormalP3uiv(tab, NAME(NormalP3uiv));
   SET_PrimitiveRestartNV(tab, NAME(PrimitiveRestartNV));
   SET_SecondaryColor3b(tab, NAME(SecondaryColor3b));
   SET_SecondaryColor3bv(tab, NAME(SecondaryColor3bv));
   SET_SecondaryColor3d(tab, NAME(SecondaryColor3d));
   SET_SecondaryColor3dv(tab, NAME(SecondaryColor3dv));
   SET_SecondaryColor3fEXT(tab, NAME(SecondaryColor3fEXT));
   SET_SecondaryColor3fvEXT(tab, NAME(SecondaryColor3fvEXT));
   SET_SecondaryColor3hNV(tab, NAME(SecondaryColor3hNV));
   SET_SecondaryColor3hvNV(tab, NAME(SecondaryColor3hvNV));
   SET_SecondaryColor3i(tab, NAME(SecondaryColor3i));
   SET_SecondaryColor3iv(tab, NAME(SecondaryColor3iv));
   SET_SecondaryColor3s(tab, NAME(SecondaryColor3s));
   SET_SecondaryColor3sv(tab, NAME(SecondaryColor3sv));
   SET_SecondaryColor3ub(tab, NAME(SecondaryColor3ub));
   SET_SecondaryColor3ubv(tab, NAME(SecondaryColor3ubv));
   SET_SecondaryColor3ui(tab, NAME(SecondaryColor3ui));
   SET_SecondaryColor3uiv(tab, NAME(SecondaryColor3uiv));
   SET_SecondaryColor3us(tab, NAME(SecondaryColor3us));
   SET_SecondaryColor3usv(tab, NAME(SecondaryColor3usv));
   SET_SecondaryColorP3ui(tab, NAME(SecondaryColorP3ui));
   SET_SecondaryColorP3uiv(tab, NAME(SecondaryColorP3uiv));
   SET_TexCoord1d(tab, NAME(TexCoord1d));
   SET_TexCoord1dv(tab, NAME(TexCoord1dv));
   SET_TexCoord1f(tab, NAME(TexCoord1f));
   SET_TexCoord1fv(tab, NAME(TexCoord1fv));
   SET_TexCoord1hNV(tab, NAME(TexCoord1hNV));
   SET_TexCoord1hvNV(tab, NAME(TexCoord1hvNV));
   SET_TexCoord1i(tab, NAME(TexCoord1i));
   SET_TexCoord1iv(tab, NAME(TexCoord1iv));
   SET_TexCoord1s(tab, NAME(TexCoord1s));
   SET_TexCoord1sv(tab, NAME(TexCoord1sv));
   SET_TexCoord2d(tab, NAME(TexCoord2d));
   SET_TexCoord2dv(tab, NAME(TexCoord2dv));
   SET_TexCoord2f(tab, NAME(TexCoord2f));
   SET_TexCoord2fv(tab, NAME(TexCoord2fv));
   SET_TexCoord2hNV(tab, NAME(TexCoord2hNV));
   SET_TexCoord2hvNV(tab, NAME(TexCoord2hvNV));
   SET_TexCoord2i(tab, NAME(TexCoord2i));
   SET_TexCoord2iv(tab, NAME(TexCoord2iv));
   SET_TexCoord2s(tab, NAME(TexCoord2s));
   SET_TexCoord2sv(tab, NAME(TexCoord2sv));
   SET_TexCoord3d(tab, NAME(TexCoord3d));
   SET_TexCoord3dv(tab, NAME(TexCoord3dv));
   SET_TexCoord3f(tab, NAME(TexCoord3f));
   SET_TexCoord3fv(tab, NAME(TexCoord3fv));
   SET_TexCoord3hNV(tab, NAME(TexCoord3hNV));
   SET_TexCoord3hvNV(tab, NAME(TexCoord3hvNV));
   SET_TexCoord3i(tab, NAME(TexCoord3i));
   SET_TexCoord3iv(tab, NAME(TexCoord3iv));
   SET_TexCoord3s(tab, NAME(TexCoord3s));
   SET_TexCoord3sv(tab, NAME(TexCoord3sv));
   SET_TexCoord4d(tab, NAME(TexCoord4d));
   SET_TexCoord4dv(tab, NAME(TexCoord4dv));
   SET_TexCoord4f(tab, NAME(TexCoord4f));
   SET_TexCoord4fv(tab, NAME(TexCoord4fv));
   SET_TexCoord4hNV(tab, NAME(TexCoord4hNV));
   SET_TexCoord4hvNV(tab, NAME(TexCoord4hvNV));
   SET_TexCoord4i(tab, NAME(TexCoord4i));
   SET_TexCoord4iv(tab, NAME(TexCoord4iv));
   SET_TexCoord4s(tab, NAME(TexCoord4s));
   SET_TexCoord4sv(tab, NAME(TexCoord4sv));
   SET_TexCoordP1ui(tab, NAME(TexCoordP1ui));
   SET_TexCoordP1uiv(tab, NAME(TexCoordP1uiv));
   SET_TexCoordP2ui(tab, NAME(TexCoordP2ui));
   SET_TexCoordP2uiv(tab, NAME(TexCoordP2uiv));
   SET_TexCoordP3ui(tab, NAME(TexCoordP3ui));
   SET_TexCoordP3uiv(tab, NAME(TexCoordP3uiv));
   SET_TexCoordP4ui(tab, NAME(TexCoordP4ui));
   SET_TexCoordP4uiv(tab, NAME(TexCoordP4uiv));
   SET_Vertex2d(tab, NAME(Vertex2d));
   SET_Vertex2dv(tab, NAME(Vertex2dv));
   SET_Vertex2f(tab, NAME(Vertex2f));
   SET_Vertex2fv(tab, NAME(Vertex2fv));
   SET_Vertex2hNV(tab, NAME(Vertex2hNV));
   SET_Vertex2hvNV(tab, NAME(Vertex2hvNV));
   SET_Vertex2i(tab, NAME(Vertex2i));
   SET_Vertex2iv(tab, NAME(Vertex2iv));
   SET_Vertex2s(tab, NAME(Vertex2s));
   SET_Vertex2sv(tab, NAME(Vertex2sv));
   SET_Vertex3d(tab, NAME(Vertex3d));
   SET_Vertex3dv(tab, NAME(Vertex3dv));
   SET_Vertex3f(tab, NAME(Vertex3f));
   SET_Vertex3fv(tab, NAME(Vertex3fv));
   SET_Vertex3hNV(tab, NAME(Vertex3hNV));
   SET_Vertex3hvNV(tab, NAME(Vertex3hvNV));
   SET_Vertex3i(tab, NAME(Vertex3i));
   SET_Vertex3iv(tab, NAME(Vertex3iv));
   SET_Vertex3s(tab, NAME(Vertex3s));
   SET_Vertex3sv(tab, NAME(Vertex3sv));
   SET_Vertex4d(tab, NAME(Vertex4d));
   SET_Vertex4dv(tab, NAME(Vertex4dv));
   SET_Vertex4f(tab, NAME(Vertex4f));
   SET_Vertex4fv(tab, NAME(Vertex4fv));
   SET_Vertex4hNV(tab, NAME(Vertex4hNV));
   SET_Vertex4hvNV(tab, NAME(Vertex4hvNV));
   SET_Vertex4i(tab, NAME(Vertex4i));
   SET_Vertex4iv(tab, NAME(Vertex4iv));
   SET_Vertex4s(tab, NAME(Vertex4s));
   SET_Vertex4sv(tab, NAME(Vertex4sv));
   SET_VertexAttrib1dNV(tab, NAME(VertexAttrib1dNV));
   SET_VertexAttrib1dvNV(tab, NAME(VertexAttrib1dvNV));
   SET_VertexAttrib1fNV(tab, NAME(VertexAttrib1fNV));
   SET_VertexAttrib1fvNV(tab, NAME(VertexAttrib1fvNV));
   SET_VertexAttrib1hNV(tab, NAME(VertexAttrib1hNV));
   SET_VertexAttrib1hvNV(tab, NAME(VertexAttrib1hvNV));
   SET_VertexAttrib1sNV(tab, NAME(VertexAttrib1sNV));
   SET_VertexAttrib1svNV(tab, NAME(VertexAttrib1svNV));
   SET_VertexAttrib2dNV(tab, NAME(VertexAttrib2dNV));
   SET_VertexAttrib2dvNV(tab, NAME(VertexAttrib2dvNV));
   SET_VertexAttrib2fNV(tab, NAME(VertexAttrib2fNV));
   SET_VertexAttrib2fvNV(tab, NAME(VertexAttrib2fvNV));
   SET_VertexAttrib2hNV(tab, NAME(VertexAttrib2hNV));
   SET_VertexAttrib2hvNV(tab, NAME(VertexAttrib2hvNV));
   SET_VertexAttrib2sNV(tab, NAME(VertexAttrib2sNV));
   SET_VertexAttrib2svNV(tab, NAME(VertexAttrib2svNV));
   SET_VertexAttrib3dNV(tab, NAME(VertexAttrib3dNV));
   SET_VertexAttrib3dvNV(tab, NAME(VertexAttrib3dvNV));
   SET_VertexAttrib3fNV(tab, NAME(VertexAttrib3fNV));
   SET_VertexAttrib3fvNV(tab, NAME(VertexAttrib3fvNV));
   SET_VertexAttrib3hNV(tab, NAME(VertexAttrib3hNV));
   SET_VertexAttrib3hvNV(tab, NAME(VertexAttrib3hvNV));
   SET_VertexAttrib3sNV(tab, NAME(VertexAttrib3sNV));
   SET_VertexAttrib3svNV(tab, NAME(VertexAttrib3svNV));
   SET_VertexAttrib4dNV(tab, NAME(VertexAttrib4dNV));
   SET_VertexAttrib4dvNV(tab, NAME(VertexAttrib4dvNV));
   SET_VertexAttrib4fNV(tab, NAME(VertexAttrib4fNV));
   SET_VertexAttrib4fvNV(tab, NAME(VertexAttrib4fvNV));
   SET_VertexAttrib4hNV(tab, NAME(VertexAttrib4hNV));
   SET_VertexAttrib4hvNV(tab, NAME(VertexAttrib4hvNV));
   SET_VertexAttrib4sNV(tab, NAME(VertexAttrib4sNV));
   SET_VertexAttrib4svNV(tab, NAME(VertexAttrib4svNV));
   SET_VertexAttrib4ubNV(tab, NAME(VertexAttrib4ubNV));
   SET_VertexAttrib4ubvNV(tab, NAME(VertexAttrib4ubvNV));
   SET_VertexAttribs1dvNV(tab, NAME(VertexAttribs1dvNV));
   SET_VertexAttribs1fvNV(tab, NAME(VertexAttribs1fvNV));
   SET_VertexAttribs1hvNV(tab, NAME(VertexAttribs1hvNV));
   SET_VertexAttribs1svNV(tab, NAME(VertexAttribs1svNV));
   SET_VertexAttribs2dvNV(tab, NAME(VertexAttribs2dvNV));
   SET_VertexAttribs2fvNV(tab, NAME(VertexAttribs2fvNV));
   SET_VertexAttribs2hvNV(tab, NAME(VertexAttribs2hvNV));
   SET_VertexAttribs2svNV(tab, NAME(VertexAttribs2svNV));
   SET_VertexAttribs3dvNV(tab, NAME(VertexAttribs3dvNV));
   SET_VertexAttribs3fvNV(tab, NAME(VertexAttribs3fvNV));
   SET_VertexAttribs3hvNV(tab, NAME(VertexAttribs3hvNV));
   SET_VertexAttribs3svNV(tab, NAME(VertexAttribs3svNV));
   SET_VertexAttribs4dvNV(tab, NAME(VertexAttribs4dvNV));
   SET_VertexAttribs4fvNV(tab, NAME(VertexAttribs4fvNV));
   SET_VertexAttribs4hvNV(tab, NAME(VertexAttribs4hvNV));
   SET_VertexAttribs4svNV(tab, NAME(VertexAttribs4svNV));
   SET_VertexAttribs4ubvNV(tab, NAME(VertexAttribs4ubvNV));
   SET_VertexP2ui(tab, NAME(VertexP2ui));
   SET_VertexP2uiv(tab, NAME(VertexP2uiv));
   SET_VertexP3ui(tab, NAME(VertexP3ui));
   SET_VertexP3uiv(tab, NAME(VertexP3uiv));
   SET_VertexP4ui(tab, NAME(VertexP4ui));
   SET_VertexP4uiv(tab, NAME(VertexP4uiv));
}
if (ctx->API == API_OPENGL_COMPAT || ctx->API == API_OPENGLES) {
   SET_Color4f(tab, NAME(Color4f));
   SET_Color4ub(tab, NAME(Color4ub));
   SET_Materialf(tab, NAME(Materialf));
   SET_Materialfv(tab, NAME(Materialfv));
   SET_MultiTexCoord4fARB(tab, NAME(MultiTexCoord4fARB));
   SET_Normal3f(tab, NAME(Normal3f));
}
if (ctx->API == API_OPENGL_COMPAT || ctx->API == API_OPENGL_CORE) {
   SET_VertexAttribL1d(tab, NAME(VertexAttribL1d));
   SET_VertexAttribL1dv(tab, NAME(VertexAttribL1dv));
   SET_VertexAttribL2d(tab, NAME(VertexAttribL2d));
   SET_VertexAttribL2dv(tab, NAME(VertexAttribL2dv));
   SET_VertexAttribL3d(tab, NAME(VertexAttribL3d));
   SET_VertexAttribL3dv(tab, NAME(VertexAttribL3dv));
   SET_VertexAttribL4d(tab, NAME(VertexAttribL4d));
   SET_VertexAttribL4dv(tab, NAME(VertexAttribL4dv));
}
